#!/usr/bin/env ruby
#--
# kate: remove-trailing-space on; replace-trailing-space-save on; indent-width 2; indent-mode ruby; syntax ruby;
#++
# The Gnuplot Ruby Module allows easy accessing the diagram plot features of
# Gnuplot in an object-oriented rubish way. Creating of images and PDF files is
# supported as output via Gnuplot windows of different types, too.
#
# Author::      Robert Riemann
# Copyright::   Copyright (c) 2010 Robert Riemann
# License::     Distributes under the same terms as Ruby

#--

class Object

  def instance_variable_not_empty? varname
    if self.instance_variable_defined? varname
      var = self.instance_variable_get varname
      content = var.to_gnuplot
      return content unless content.empty?
    end
    nil
  end

  def is_bool?
    self == false or self == nil or self == true
  end

  def to_gnuplot
    to_s
  end

end

class Range

  def to_gnuplot
    "#{self.first}:#{self.last}"
  end

end

# class String
#   alias to_gnuplot to_s
# end
#
# class Symbol
#   alias to_gnuplot to_s
# end
#
# class Fixnum
#   alias to_gnuplot to_s
# end
#
# class FalseClass
#   alias to_gnuplot to_s
# end
#
# class TrueClass
#   alias to_gnuplot to_s
# end

#++

class Array

  def to_gnuplot_data
    if self[0].kind_of? Array
      self[0].zip( *self[1..-1] ).collect { |a| a.join(" ") }.join("\n") + "\ne"
    elsif self[0].kind_of? Numeric
      s = ""
      self.length.times { |i| s << "#{self[i]}\n" }
      s
    else
      self[0].zip( *self[1..-1] ).to_gplot
    end
  end

  def childs_are? a_class
    self.each do |child|
      raise "#{child} needs to be a #{a_class.to_s}." unless child.is_a? a_class
    end
  end

  def self.new_first_order object
    if object.is_a? Array
      object
    else
      [object]
    end
  end

end

module Gnuplot


  ##
  # The Plot class represents the \Gnuplot _plot_ command.
  #
  # Example:
  #  Gnuplot::Plot.new(graphs: singraph, title: "plot title", xlabel: "xlabel", ylabel: "ylabel", ranges: 0..6, samples: 25)
  class Plot

    # QUOTED contains all \Gnuplot attributes which expects a _String_ in quotes.
    # The quotes will be added automatically, if not already present.
    QUOTED = [ "title", "output", "xlabel", "ylabel" ]

    # defines the \Gnuplot command to "plot"
    CMD = "plot"

    attr_reader :graphs, :ranges, :sets

    def initialize options

      self.graphs = (options.include? :graphs) ? options.delete( :graphs ) : []
      self.ranges = (options.include? :ranges) ? options.delete( :ranges ) : []

      @sets = Hash.new
      options.each do |attr,value|
        set attr.to_s, *value
      end

    end

    def graphs= graphs
      graphs = Array.new_first_order graphs
      graphs.childs_are? Graph
      @graphs = graphs
    end

    def ranges= ranges
      ranges = Array.new_first_order ranges
      ranges.childs_are? Range
      @ranges = ranges
    end

    def method_missing attribute, *arguments
      attribute = attribute.to_s
      attribute = attribute[0..-2] if attribute[-1] == "="
      set attribute, *arguments
    end

    def set attribute, *arguments
      if QUOTED.include? attribute
        raise "#{attribute} needs 1 argument" if arguments.size == 0
        arguments[0] = "'#{arguments[0]}'" unless arguments[0] =~ /^('|").*\1$/
        arguments = [arguments[0]]
      else
        arguments = arguments[0] if arguments[0].is_bool?
      end
      @sets[ attribute ] = arguments
    end

    def to_gnuplot
      STDERR.puts @sets.inspect
      return "" if @graphs.empty?

      str = ""

      @sets.each do |attribute, values|
        if values.is_a? Array
          values.each do |value|
            str += "set #{attribute} #{value.to_gnuplot}\n"
          end
        else
          str += values ? "unset #{attribute}\n" : "set #{attribute}\n"
        end
      end

      str += self.class::CMD + " "

      @ranges.each do |range|
        str += "[#{range.to_gnuplot}] "
      end

      @graphs.each_index do |i|
        str += ", " unless i == 0
        str += @graphs[i].to_gnuplot
      end

      str += "\n"

      @graphs.each do |graph|
        if graph.is_a? DataGraph and graph.data_included?
          str += graph.to_gnuplot_data
        end
      end

      str
    end

  end

  ##
  # The SPlot class represents the \Gnuplot _splot_ command.
  #
  # Example:
  #  Gnuplot::SPlot.new(graphs: sincosgraph, title: "plot title", xlabel: "xlabel", ylabel: "ylabel", ranges: 0..6, samples: 25)
  #
  # If you want to be able to rotate your surface plots in your plot window,
  # have to make sure to not close the Instance.
  class SPlot < Plot

    # defines the \Gnuplot command to "splot"
    CMD = "splot"

  end

  class Graph

    ##
    # This method gives access to the title displayed in the legend of the plot.
    # It expects a _String_.
    #
    # If you want to get the \Gnuplot default title, just don't set the title or
    # set the title to nil or "". Make sure to don't have #notitle set to true.
    #
    # Example:
    #
    #  graph_object.title = "custom title"
    attr_accessor :title

    ##
    # If you don't want to get any entry in the plot legend at all, you have to
    # set the notitle attribute to true. This method expects a _Boolean_. If it
    # is set to true, the value of #title won't have no influence.
    attr_accessor :notitle

    ##
    # You can change the look of your graph by applying a Style.
    #
    # Example:
    #
    #  graph_object.style = style_object
    attr_accessor :style

    attr_accessor :using

    attr_accessor :prepend

    ##
    # The constructor accepts a Hash containing the single attributes.
    # So Example 1 and 2 are both equivalent code.
    #
    # Example 1
    #
    #  formulagraph_object = FormulaGraph.new
    #  formulagraph_object.formula = "x**2"
    #
    # Example 2
    #
    #  formulagraph_object = FormulaGraph.new( formula: "x**2" )
    def initialize options
      options.each do |attr,value|
        attr = ("@" + attr.to_s)
        instance_variable_set attr, value
      end
    end

    ##
    # This method retuns the _String_ containing the gnuplot code. It gets used
    # along with Plot#to_gnuplot.
    def to_gnuplot
      str = ""

      if instance_variable_not_empty? "@notitle"
        str += "notitle "
      else
        s = instance_variable_not_empty? "@title"
        str += "title \"#{s}\" " if s
      end

      s = instance_variable_not_empty? "@using"
      str += "using #{s} " if s

      s = instance_variable_not_empty? "@style"
      str += s if s

      s = instance_variable_not_empty? "@prepend"
      str += s if s

      str
    end

  end


  ##
  # Use a FormulaGraph to plot a formula with Plot. Make sure to set the
  # #formula attribute as it is mendatory.
  class FormulaGraph < Graph

    ##
    # This method allows to read and set the formula. It expects a _String_.
    #
    # The formula will be used by \Gnuplot to generate the function values. All
    # functions predefined by \Gnuplot can be used. For more information, take a
    # look at the {Gnuplot manual}[http://www.gnuplot.info/docs/gnuplot.html#x1-5300013.1].
    # Please make sure to pass valid \Gnuplot commands.
    #
    # Example:
    #
    #  formulargraph_object.formula = "sin(x)"
    #
    attr_accessor :formula

    ##
    # This method retuns the _String_ containing the gnuplot code. It gets used
    # along with Plot#to_gnuplot.
    def to_gnuplot
      str = ""

      s = instance_variable_not_empty?( "@formula" ) or raise "formula wasn't set."
      str += s + " "

      str += super

      str
    end

  end

  class DataGraph < Graph

    attr_accessor :data
    attr_reader :data_included
    alias data_included? data_included

    ##
    # This method retuns the _String_ containing the gnuplot code. It gets used
    # along with Plot#to_gnuplot.
    def to_gnuplot
      str = ""

      raise "data is in DataGraph mendatory and wasn't set." unless instance_variable_defined? "@data"
      if @data.is_a? String
        STDERR.puts "Warning: couldn't read file #{@data}." if File.readable? @data
        str += "'#{@data}' "
        @data_included = nil
      else
        str += "'-' "
        @data_included = true
      end

      str += super

      str
    end


    def to_gnuplot_data
      @data.to_gnuplot_data
    end

  end

  ##
  # Use a Style along with a Graph object to control the appearance in the plot.
  #
  # Example:
  #
  #  style_object = Style.new(color: :gold)
  #  graph_object = FormulaGraph.new(formula: "sin(x)", style: style_object)
  class Style

    ##
    # This attributes controls which kind of plotting style should be used.
    # This attributes expects a _String_ or a _Symbol_.

    # Possible values are: :lines, "points", :dots, etc.
    # More options are documented in the {Gnuplot manual}[http://www.gnuplot.info/docs/gnuplot.html#x1-14700034.7].
    attr_accessor :type

    ##
    # Sets point size, expects _Fixnum_
    attr_accessor :size

    ##
    # Sets line width, expects _Fixnum_
    attr_accessor :width

    ##
    # To set the color of your graph you have two options.
    # - color is a _Fixnum_: A number, which represents a preset depending on the
    #   used terminal will be used to set the color.
    # - color is a _String_ or a _Symbol_: Color will be interpreted as a color name
    #   wich has to be supported by the given terminal.
    #   \Gnuplot code will be: rgb 'color'
    attr_accessor :color

    def initialize options
      options.each do |attr,value|
        attr = ("@" + attr.to_s)
        instance_variable_set attr, value
      end
    end

    def to_gnuplot
      str = ""

      s = instance_variable_not_empty? "@type"
      str += "with #{s} " if s

      s = instance_variable_not_empty? "@color"
      str += (s =~ /^\d+$/) ? "linecolor #{s} " : "linecolor rgb \'#{s}\' " if s

      s = instance_variable_not_empty? "@size"
      str += "pointsize #{s} " if s

      s = instance_variable_not_empty? "@width"
      str += "linewidth #{s} " if s

      str
    end

  end

  def Gnuplot.terminals
    terms = nil
    Gnuplot::Instance.open( nil, "r+" ) do |instance|
      begin
        instance.puts "set terminal"
        instance.close_write
        terms = instance.read.split($/).map do |line|
          m = /^(?!Available terminal types:)\s*(\S+)\s+(.*)\s*$/.match(line)
          m[1..-1] if m
        end - [nil]
      rescue IOError
        STDERR.puts "Could not get Gnuplot terminal types"
      end
    end
    terms
  end

  def Gnuplot.which ( bin )
    return Gnuplot::GNUPLOT_EXE if const_defined? "GNUPLOT_EXE"
    return bin if File::executable? bin

    if RUBY_PLATFORM =~ /linux/i
      candidate = `which #{bin}`.chomp
    else
      path = ENV['PATH'] # || ENV['WHAT_EVER_WINDOWS_PATH_VAR_IS']
      path.split(File::PATH_SEPARATOR).each do |dir|
        candidate = File::join dir, bin.strip
      end
    end

    if File::executable? candidate
      const_set "GNUPLOT_EXE", candidate
      candidate
    else
      nil
    end
  end

  class Instance < IO

    def self.open( persist = true, mode = "w" )
      cmd = Gnuplot.which( ENV['RB_GNUPLOT'] || 'gnuplot' ) or raise 'gnuplot not found'
      STDERR.puts "Found " + cmd + "\n"
      cmd += " --persist" if persist
      io = self.popen( cmd, mode)
      if block_given?
         yield io
         io.close
      end
      io
    end

  end

end

# sinstyle = Gnuplot::Style.new(color: :green, type: :fsteps, width: 5)
# singraph = Gnuplot::FormulaGraph.new(formula: "sin(x)", title: "der Sinus", style: sinstyle)
# x = (0..20).collect { |v| v.to_f/2 }
# y = x.collect { |v| Math.cos v }
# datgraph = Gnuplot::DataGraph.new(data: [x,y], title: "data plot", style: Gnuplot::Style.new(color: :red, type: :lines))
# sincosgraph = Gnuplot::FormulaGraph.new(formula: "sin(x)*abs(sqrt(y*5))", title: "der Sinus*Kosinus", style: sinstyle)
# sinplot = Gnuplot::Plot.new(graphs: [singraph, datgraph], title: "Der Sinus-Plot und der Dat-Plot", xlabel: "x Beschriftung", ylabel: "y Beschriftung")
# sinplot = Gnuplot::SPlot.new(graphs: sincosgraph)
# data = [[1, 2, 3],[1, 2, 3]]
graph = Gnuplot::DataGraph.new data: "immigration.dat", using: "6:xtic(1) ti col", prepend: ", '' u 12 ti col, '' u 13 ti col, '' u 14 ti col"
plot = Gnuplot::Plot.new graphs: graph, boxwidth: "0.9 relative", style: ["data histogram", "fill solid 1.0"]
Gnuplot::Instance.open true do |gp|
  gp.puts "set terminal x11"
  s = plot.to_gnuplot
  STDERR.puts s
  gp.puts s
end